__author__ = "Robert C. Kirby (kirby@uchicago.edu)"
__date__ = "2005-04-01 -- 2006-04-01"
__copyright__ = "Copyright (C) 2006- Robert C. Kirby"
__license__  = "GNU LGPL Version 2"

import sigdig, Numeric, string, graph
from xpermutations import xuniqueCombinations

def unit_vector( v ):
    return sigdig.vec_round_sig( v / max( abs( v ) ) , 4 )

def nnz( u ):
    zs = [ u[0] - u[0] ] * len(u)
    return len(u) - sum( zs == u )

def edplus( u , v ):
    ur = sigdig.vec_round_sig( u , 5 )
    vr = sigdig.vec_round_sig( v , 5 )
    return nnz( ur - vr )

def edminus( u , v ):
    ur = sigdig.vec_round_sig( u , 5 )
    vr = sigdig.vec_round_sig( v , 5 )
    return nnz( ur + vr )


def colinear( u , v ):
    d = len( u )
    z = Numeric.zeros( u.shape , "d" )
    uzhuh = Numeric.alltrue( Numeric.allclose( u , z ) )
    vzhuh = Numeric.alltrue( Numeric.allclose( v , z ) )
    if uzhuh and vzhuh:
        return 0
    elif uzhuh or vzhuh:
        return d
    else:
        uhat = unit_vector( u )
        vhat = unit_vector( v )
        if Numeric.alltrue( Numeric.allclose( uhat , vhat ) ) \
           or Numeric.alltrue( Numeric.allclose( uhat , -vhat ) ):
            return 1
        else:
            return d

def rho( u , v ):
    ep = edplus( u , v )
    em = edminus( u , v )
    c = colinear( u , v )
    m = min( em , ep , c )
    if ep == m:
        l = "edp"
    elif em == m:
        l = "edm"
    else:
        l = "c"
    return (m,l)

def get_graph( vecs , dist ):
    G = dict(  [ (v,{}) for v in vecs ] )
    for (u,v) in xuniqueCombinations( vecs.keys() , 2 ):
        (w,l) = dist( vecs[u] , vecs[v] )
        G[u][v] = (w,l)
        G[v][u] = (w,l)

    return G

def process_bf( vecs ):
    return graph.prim( get_graph( vecs , rho ) )

def cost( g , vecs ):
    """Takes a MST graph constructed in process_bf and returns
    the sum of edge weights plus the number of nonzeros in the root;
    this is the total MAPs needed to form an element stiffness matrix."""
    w = 0
    for u in g:
        if g[u]:
            v = g[u].keys()[0]
            w += g[u][v][0]
        else:
            w += nnz( vecs[u] )
    return w

def nodep_code( u ):
    return [ (u[i],1,i) for i in range(len(u)) ]

# returns the abstract code
# associated with various dependencies
# u^t g is assumed known, v^t is to be computed.
# v^t g = u^t g + ( v - u )^t g
def ep_code( Adict , i1 , i2 ):
    u = Adict[i1]
    v = Adict[i2]
    diff = v - u
    diffinds = []
    # which indices matter?
    for i in range( len( diff ) ):
        if abs( diff[i] ) >= 1.e-4:
            diffinds.append(i)

    oplist = [ (1.0,0,i1) ]
    for i in diffinds:
        oplist.append( (diff[i],1,i) )

    return oplist

# returns the abstract code
# associated with various dependencies
# u^t g is assumed known, v^t is to be computed.
# v^t g = -u^t g + ( v + u )^t g
def em_code( Adict , i1 , i2 ):
    u = Adict[i1]
    v = Adict[i2]
    diff = v + u
    diffinds = []
    # which indices matter?
    for i in range( len( diff ) ):
        if abs( diff[i] ) >= 1.e-4:
            diffinds.append(i)
	    
    oplist = [ (-1.0,0,i1) ]
    for i in diffinds:
        oplist.append( (diff[i],1,i) )

    return oplist



# v^t g = v[0]/u[0] ( u^t g )
def c_code( Adict , i1 , i2 ):
    u = Adict[i1]
    v = Adict[i2]
    i=0
    while abs(u[i]) < 1.e-6 and i<len(u):
	i+=1
    
    alpha = v[i] / u[i]

    return [ ( alpha , 0 , i1 ) ] 

def abstract_code( mst , Adict ):
    """Takes the MST as input and returns a list modeling
    primitive abstract syntax for the optimized code """
    ts = graph.topsort( mst )
    ops = []
    for k in ts:
        if mst[k]:
            if len( mst[k] ) == 1:
                p = mst[k].iterkeys().next()
                if mst[k][p][1] == 'c':
                    ops.append( ( k , c_code( Adict , p , k ) ) )
                elif mst[k][p][1] == 'edp':
                    ops.append( ( k , ep_code( Adict , p , k ) ) )
                elif mst[k][p][1] == 'edm':
                    ops.append( ( k , em_code( Adict , p , k ) ) )
                else:
                    ops.append( ( k , nodep_code(Adict[k]) ) )
                
        else:
            ops.append( ( k , nodep_code( Adict[k] ) ) )
    return ops

def opt_code( A0 , p , Adict ):
    if len( Adict.keys()[0] ) != 2:
	raise RuntimeError, "Illegal input"
    cs = abstract_code( p , Adict )
    code_list = []
    def flatten(iota):
	return iota[0]*A0.shape[0] + iota[1]

    def convert( a ):
	if a[1] == 0:
	    return (a[0],a[1],flatten(a[2]))
	else:
	    return a

    for c in cs:
	lvalue = (0,flatten(c[0]))
	rvalue = map( convert , c[1] )
	code_list.append( (lvalue,rvalue) )
    return code_list

def optimize( A0 ):
    Adict = {}
    for i in range( A0.shape[0] ):
	for j in range( A0.shape[1] ):
	    Adict[i,j] = Numeric.reshape( A0[i,j] , (-1,) )
    p = process_bf( Adict )
    return opt_code(A0,p,Adict )


def main():
    import build_tensors
    shape = "triangle"
    degree = 2
    A0 = build_tensors.laplacianform(shape,degree)
    for c in optimize( A0 ):
	print c
    
def mainbak():
    import build_tensors
    for shape in ("triangle",):
        for degree in range(2,3):
            Adict = build_tensors.laplacian( shape , degree )
            ks = Adict.keys()
            ks.sort()
            p = process_bf( Adict )
            print p
            sum_pwts = 0
            for u in p:
                for v in p[u]:
                    sum_pwts += p[u][v][0]
            wt = len( Adict.itervalues().next() ) + sum_pwts
            print len(Adict)*len(Adict.itervalues().next()),wt
	    print string.join( opt_code( p , Adict ) , "\n" )
#            dot = graph.Dot( p )
#            for u in p:
#                for v in p[u]:
#                    dot.set_edge_style( u , v , label = '%s' % (p[u][v][0]) )
#            dot.save_image( "foo.gif" )
##             cs = abstract_code( p , Adict )
##             for c in cs:
##                 code = string.join( [ "%s * %s[%s]"  \
##                                       % foo for foo in c[1] ] , " + " )
##                 print "K[%s] = %s " % ( c[0] , code )


if __name__ == "__main__":
    main()
    #import profile
    #import pstats
    #profile.run( 'main()' , 'fooprof')
    #p = pstats.Stats('fooprof')
    #p.sort_stats('time','cum')
    #p.print_stats()
